unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls;

const
  AM_GOTDATA = WM_USER + $99;

type
  TMainForm = class(TForm)
    DeviceGroup: TGroupBox;
    DeviceCombo: TComboBox;
    DeviceDetectButton: TButton;
    LogGroup: TGroupBox;
    TimestampRadio: TRadioButton;
    GivenNameRadio: TRadioButton;
    GivenNameEdit: TEdit;
    AD0RangeRadios: TRadioGroup;
    Panel1: TPanel;
    StartButton: TButton;
    StopButton: TButton;
    LogMemo: TMemo;
    Label1: TLabel;
    AD1RangeRadios: TRadioGroup;
    NoLogRadio: TRadioButton;
    TimestampLabel: TLabel;
    ADCLabel0: TLabel;
    ADCLabel1: TLabel;
    cbLogNpy: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure ADRangeRadiosClick(Sender: TObject);
    procedure GivenNameEditChange(Sender: TObject);
    procedure StartButtonClick(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
    procedure DeviceDetectButtonClick(Sender: TObject);
    procedure DeviceComboChange(Sender: TObject);
    procedure cbLogNpyClick(Sender: TObject);
  protected
    DeviceIndex: LongWord;

    procedure AMGotData(var Msg: TMessage); message AM_GOTDATA;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  Registry, Consts,
  AIOUSB,
  LogThreads;



const
  MyKey = 'Software\ACCES\Logger\';
  RangeValueName: array[0..1] of String = ('AD0Range', 'AD1Range');

var
  LogThread: TLogThread;
  LeftoverData: AnsiString;
  Config: array[0..1] of Byte;
  LastTimestamp: LongWord;
  LastV: array[0..1] of Double;
  extLogFile: AnsiString;
  bNpyLog: Boolean;

type
  TUSBAI162APoint = packed record
    ADC0, ADC1: Word;
    DITS: LongWord;
  end;
  TPUSBAI162APoint = ^TUSBAI162APoint;

procedure ADCallback(pBuf: PWord; BufSize, Flags, Context: LongWord); cdecl; forward;

function MultiStrCat(const Strs: array of String): String;
var
  iStr, iCh, L: Integer;

begin
  L := 0;
  for iStr := 0 to High(Strs) do
    Inc(L, Length(Strs[iStr]))
  ;
  SetString(Result, nil, L);

  iCh := 1;
  for iStr := 0 to High(Strs) do begin
    L := Length(Strs[iStr]);
    Move(Strs[iStr][1], Result[iCh], L);
    Inc(iCh, L);
  end;
end;

procedure EnabledWithChildren(WinControl: TWinControl; NewEnabled: Boolean);
var
  I: Integer;

begin
  WinControl.Enabled := NewEnabled;
  for I := 0 to WinControl.ControlCount-1 do
    WinControl.Controls[I].Enabled := NewEnabled
  ;
end;

function VoltsFromCounts(Ct: Word; RangeCode: Byte): Double;
begin
  Result := 1/65536 * Ct;
  if (RangeCode and 1) <> 0 then Result := Result * 2 - 1;
  if (RangeCode and 2) = 0 then Result := Result * 2;
  if (RangeCode and 4) = 0 then Result := Result * 5;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  Reg: TRegistry;

begin
  IsMultiThread := True;

  Reg := TRegistry.Create(KEY_READ);
  Reg.RootKey := HKEY_CURRENT_USER;
  if Reg.OpenKey(MyKey, False) then begin
    try AD0RangeRadios.ItemIndex := Reg.ReadInteger(RangeValueName[0]); except AD0RangeRadios.ItemIndex := 1; end;
    try AD1RangeRadios.ItemIndex := Reg.ReadInteger(RangeValueName[1]); except AD1RangeRadios.ItemIndex := 1; end;
    Reg.CloseKey;
  end;
  Reg.Free;

  AD0RangeRadios.OnClick := ADRangeRadiosClick;
  AD1RangeRadios.OnClick := ADRangeRadiosClick;

  DeviceDetectButton.Click;
  cbLogNpyClick(nil);
end;

procedure TMainForm.ADRangeRadiosClick(Sender: TObject);
var
  Reg: TRegistry;

begin
  Reg := TRegistry.Create(KEY_READ or KEY_WRITE);
  Reg.RootKey := HKEY_CURRENT_USER;
  if Reg.OpenKey(MyKey, True) then begin
    try Reg.DeleteValue(RangeValueName[0]); except end;
    try Reg.WriteInteger(RangeValueName[0], AD0RangeRadios.ItemIndex); except end;
    try Reg.DeleteValue(RangeValueName[1]); except end;
    try Reg.WriteInteger(RangeValueName[1], AD1RangeRadios.ItemIndex); except end;
    Reg.CloseKey;
  end;
  Reg.Free;
end;

procedure TMainForm.GivenNameEditChange(Sender: TObject);
begin
  GivenNameRadio.Checked := True;
end;

procedure TMainForm.DeviceDetectButtonClick(Sender: TObject);
var
  iDeviceIndex, MaskDevices, Status, PID, DeviceNameSize: LongWord;
  DeviceName: AnsiString;

begin
  DeviceIndex := diNone;
  DeviceCombo.Items.Clear;

  for iDeviceIndex := 0 to 31 do AIOUSB_CloseDevice(iDeviceIndex);
  MaskDevices := GetDevices;
  for iDeviceIndex := 0 to 31 do
    if (MaskDevices and (1 shl iDeviceIndex)) <> 0 then begin
      DeviceNameSize := 256;
      SetString(DeviceName, nil, DeviceNameSize);
      Status := QueryDeviceInfo(iDeviceIndex, @PID, @DeviceNameSize, @DeviceName[1], nil, nil);
      if Status <> ERROR_SUCCESS then Continue;

      if PID <> $8180 then Continue; // Not what we're looking for.

      SetLength(DeviceName, DeviceNameSize);

      DeviceCombo.Items.Add('[' + IntToStr(iDeviceIndex) + '] ' + DeviceName + '(' + IntToHex(PID, 4) + ')');
    end
  ;
  if DeviceCombo.Items.Count <> 0 then begin
    DeviceCombo.ItemIndex := 0;
    DeviceComboChange(DeviceCombo);
  end;
end;

procedure TMainForm.StartButtonClick(Sender: TObject);
var
  Status: LongWord;
  LogFN: String;
  ST: TSystemTime;
  ConfigSize: LongWord;
  BufSize, BufCount: LongWord;
const
  npyHeaderStuff: AnsiString=#93'NUMPY'#01#00#54#00'{"descr": "[("CH0", "<u2"), ("CH1", "<u2"), ("Timestamp", "<u4")]", "shape": (2, 2), }';

begin
  LogMemo.Lines.Add('Starting');
  bNpyLog := cbLogNpy.checked;
  if TimestampRadio.Checked then begin
    GetLocalTime(ST);
    with ST do LogFN := Format('Log %.4d%.2d%.2d_%.2d%.2d%.2d.', [ wYear, wMonth, wDay, wHour, wMinute, wSecond ]);
    LogFN := LogFN + extLogFile;
  end
  else if GivenNameRadio.Checked then begin
    LogFN := GivenNameEdit.Text;
  end
  else begin
    LogFN := '';
  end;

  if LogFN = '' then begin
    LogThread := nil;
  end
  else begin
    LogThread := TLogThread.Create(LogFN);
    if bNpyLog then
        // LogThread.WriteStr( npyHeaderStuff + StringOfChar(' ', 64-length(npyHeaderStuff)))
    else
        LogThread.WriteStr('"s","A/D 0","A/D 1"'#13#10);
  end;

  Config[0] := AD0RangeRadios.ItemIndex;
  Config[1] := AD1RangeRadios.ItemIndex;
  ConfigSize := Length(Config);
  Status := ADC_SetConfig(DeviceIndex, @Config[0], ConfigSize);
  if Status <> ERROR_SUCCESS then begin
    LogMemo.Lines.Add('Error ' + IntToStr(Status) + ' from ADC_SetConfig.');
    Exit;
  end;

  AIOUSB_ClearFIFO(DeviceIndex, 0);
  LeftoverData := '';

  BufSize := 1024*1024;
  BufCount := 64;
  Status := ADC_BulkContinuousCallbackStart(DeviceIndex, BufSize, BufCount, 0, @ADCallback);
  if Status <> ERROR_SUCCESS then begin
    LogMemo.Lines.Add('Error ' + IntToStr(Status) + ' from ADC_BulkContinuousCallbackStart.');
    Exit;
  end;

  LogMemo.Lines.Add('Started.');

  StopButton.Enabled := True;
  StartButton.Enabled := False;
  AD0RangeRadios.Enabled := False;
  AD1RangeRadios.Enabled := False;
  EnabledWithChildren(DeviceGroup, False);
  EnabledWithChildren(LogGroup, False);
end;

procedure ADCallback(pBuf: PWord; BufSize, Flags, Context: LongWord); cdecl;
var
  pDataPoint: TPUSBAI162APoint;
  ThisData: AnsiString;
  WholePoints, iPoint: Integer;
  oL: LongWord;
  LineStr: array of String;
  usTimestamp: LongWord;
  ADC0V, ADC1V: Double;

begin
  if BufSize = 0 then Exit; // We don't care about Flags.

  oL := Length(LeftoverData);
  if oL = 0 then begin
    // Just get the new data.
    SetString(ThisData, PAnsiChar(pBuf), BufSize);
  end
  else begin
    // Append the new data onto the previous leftover data.
    SetString(ThisData, nil, oL + BufSize);
    Move(LeftoverData[1], ThisData[1], oL);
    Move(pBuf^, ThisData[1 + oL], BufSize);
  end;

  WholePoints := Length(ThisData) div SizeOf(pDataPoint^);

  pDataPoint := @ThisData[1];

  if WholePoints > 0 then begin
    // There's whole points of data, so parse it.
    if LogThread = nil then begin
      //If there's multiple points, then skip to the last one.
      if WholePoints > 1 then Inc(pDataPoint, WholePoints-1);

      // Parse the raw data.
      usTimestamp := pDataPoint.DITS shr 5;
      ADC0V := VoltsFromCounts(PDataPoint.ADC0, Config[0]);
      ADC1V := VoltsFromCounts(PDataPoint.ADC1, Config[1]);

      // Ready for display.
      LastTimestamp := usTimestamp;
      LastV[0] := ADC0V;
      LastV[1] := ADC1V;

      // Advance to the leftover data if any.
      Inc(pDataPoint);
    end
    else begin // Logging is enabled, so log these data points.
        if not bNpyLog then begin  // binary log format
              SetLength(LineStr, WholePoints);
              for iPoint := 0 to WholePoints-1 do begin
                //Parse the raw data.
                usTimestamp := pDataPoint.DITS shr 5;
                ADC0V := VoltsFromCounts(PDataPoint.ADC0, Config[0]);
                ADC1V := VoltsFromCounts(PDataPoint.ADC1, Config[1]);

                //Convert to string.
                LineStr[iPoint] := IntToStr(usTimestamp) + ','
                                 + FloatToStr(ADC0V) + ','
                                 + FloatToStr(ADC1V) + #13#10
                ;

                //Ready for display.
                LastTimestamp := usTimestamp;
                LastV[0] := ADC0V;
                LastV[1] := ADC1V;

                //Advance to the next point. (Or after the last point, to the leftover data if any.)
                Inc(pDataPoint);
                LogThread.WriteStr(MultiStrCat(LineStr));
              end;
        end
        else // npy log format enabled
        begin
              for iPoint := 0 to WholePoints-1 do begin
                //Parse the raw data.
                usTimestamp := pDataPoint.DITS shr 5;
                ADC0V := VoltsFromCounts(PDataPoint.ADC0, Config[0]);
                ADC1V := VoltsFromCounts(PDataPoint.ADC1, Config[1]);

                //Ready for display.
                LastTimestamp := usTimestamp;
                LastV[0] := ADC0V;
                LastV[1] := ADC1V;

                //Advance to the next point. (Or after the last point, to the leftover data if any.)
                Inc(pDataPoint);
              end;
              LogThread.WriteStr(ThisData);
        end;
    end;
  end;

  // Update the leftover data for next time.
  SetString(LeftoverData, PAnsiChar(pDataPoint), Length(ThisData) - WholePoints*SizeOf(pDataPoint^));

  // Notify the display that we have data ready for it.
  PostMessage(MainForm.Handle, AM_GOTDATA, 0, 0);
end;

procedure TMainForm.AMGotData(var Msg: TMessage);
begin
  Msg.Result := 0;

  TimestampLabel.Caption := 'Timestamp: ' + IntToStr(LastTimestamp) + 's';
  ADCLabel0.Caption := 'ADC 0: ' + FloatToStrF(LastV[0], ffFixed, 16, 3) + 'V';
  ADCLabel1.Caption := 'ADC 1: ' + FloatToStrF(LastV[1], ffFixed, 16, 3) + 'V';
end;

procedure TMainForm.StopButtonClick(Sender: TObject);
var
  Status, IOStatus: LongWord;

begin
  LogMemo.Lines.Add('Stopping');
  Status := ADC_BulkContinuousEnd(DeviceIndex, @IOStatus);
  if Status <> ERROR_SUCCESS then
    LogMemo.Lines.Add('Error ' + IntToStr(Status) + ' from ADC_BulkContinuousEnd.')
  else if IOStatus <> ERROR_SUCCESS then
    LogMemo.Lines.Add('I/O error ' + IntToStr(IOStatus) + ' occurred.')
  else
    LogMemo.Lines.Add('Stopped.')
  ;

  LogThread.Free;

  LogMemo.Lines.Add('Logged to ' + IntToStr(LastTimestamp) + 's.');

  StartButton.Enabled := True;
  StopButton.Enabled := False;
  AD0RangeRadios.Enabled := True;
  AD1RangeRadios.Enabled := True;
  EnabledWithChildren(DeviceGroup, True);
  EnabledWithChildren(LogGroup, True);
end;

procedure TMainForm.DeviceComboChange(Sender: TObject);
var
  I: Integer;
  DevStr: String;

begin
  DeviceIndex := diNone;

  I := DeviceCombo.ItemIndex;
  if I < 0 then Exit;

  DevStr := DeviceCombo.Items[I];
  DevStr := Copy(DevStr, 2, MAXINT); //Cut off [.
  I := Pos(']', DevStr);
  DevStr := Copy(DevStr, 1, I-1); //Cut off ] and the rest.

  DeviceIndex := StrToInt(DevStr);
end;

procedure TMainForm.cbLogNpyClick(Sender: TObject);
begin
  if cbLogNpy.checked then extLogFile := 'bin' else extLogFile := 'csv';
end;

end.
